@@ -23,11 +23,13 @@ import (
 
 	"github.com/kubernetes-csi/external-resizer/pkg/resizer"
 	"github.com/kubernetes-csi/external-resizer/pkg/util"
-
+	"google.golang.org/grpc/codes"
+	"google.golang.org/grpc/status"
 	v1 "k8s.io/api/core/v1"
-	k8serrors "k8s.io/apimachinery/pkg/api/errors"
 	"k8s.io/apimachinery/pkg/api/resource"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/types"
+	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
 	"k8s.io/apimachinery/pkg/util/wait"
 	"k8s.io/client-go/informers"
 	"k8s.io/client-go/kubernetes"
@@ -37,7 +39,7 @@ import (
 	"k8s.io/client-go/tools/cache"
 	"k8s.io/client-go/tools/record"
 	"k8s.io/client-go/util/workqueue"
-	"k8s.io/klog"
+	"k8s.io/klog/v2"
 )
 
 // ResizeController watches PVCs and checks if they are requesting an resizing operation.
@@ -53,10 +55,19 @@ type resizeController struct {
 	kubeClient    kubernetes.Interface
 	claimQueue    workqueue.RateLimitingInterface
 	eventRecorder record.EventRecorder
-	pvLister      corelisters.PersistentVolumeLister
 	pvSynced      cache.InformerSynced
-	pvcLister     corelisters.PersistentVolumeClaimLister
 	pvcSynced     cache.InformerSynced
+
+	usedPVCs *inUsePVCStore
+
+	podLister       corelisters.PodLister
+	podListerSynced cache.InformerSynced
+
+	// a cache to store PersistentVolume objects
+	volumes cache.Store
+	// a cache to store PersistentVolumeClaim objects
+	claims                 cache.Store
+	handleVolumeInUseError bool
 }
 
 // NewResizeController returns a ResizeController.
@@ -65,10 +76,11 @@ func NewResizeController(
 	resizer resizer.Resizer,
 	kubeClient kubernetes.Interface,
 	resyncPeriod time.Duration,
-	informerFactory informers.SharedInformerFactory) ResizeController {
+	informerFactory informers.SharedInformerFactory,
+	pvcRateLimiter workqueue.RateLimiter,
+	handleVolumeInUseError bool) ResizeController {
 	pvInformer := informerFactory.Core().V1().PersistentVolumes()
 	pvcInformer := informerFactory.Core().V1().PersistentVolumeClaims()
-
 	eventBroadcaster := record.NewBroadcaster()
 	eventBroadcaster.StartLogging(klog.Infof)
 	eventBroadcaster.StartRecordingToSink(&corev1.EventSinkImpl{Interface: kubeClient.CoreV1().Events(v1.NamespaceAll)})
@@ -76,18 +88,20 @@ func NewResizeController(
 		v1.EventSource{Component: fmt.Sprintf("external-resizer %s", name)})
 
 	claimQueue := workqueue.NewNamedRateLimitingQueue(
-		workqueue.DefaultControllerRateLimiter(), fmt.Sprintf("%s-pvc", name))
+		pvcRateLimiter, fmt.Sprintf("%s-pvc", name))
 
 	ctrl := &resizeController{
-		name:          name,
-		resizer:       resizer,
-		kubeClient:    kubeClient,
-		pvLister:      pvInformer.Lister(),
-		pvSynced:      pvInformer.Informer().HasSynced,
-		pvcLister:     pvcInformer.Lister(),
-		pvcSynced:     pvcInformer.Informer().HasSynced,
-		claimQueue:    claimQueue,
-		eventRecorder: eventRecorder,
+		name:                   name,
+		resizer:                resizer,
+		kubeClient:             kubeClient,
+		pvSynced:               pvInformer.Informer().HasSynced,
+		pvcSynced:              pvcInformer.Informer().HasSynced,
+		claimQueue:             claimQueue,
+		volumes:                pvInformer.Informer().GetStore(),
+		claims:                 pvcInformer.Informer().GetStore(),
+		eventRecorder:          eventRecorder,
+		usedPVCs:               newUsedPVCStore(),
+		handleVolumeInUseError: handleVolumeInUseError,
 	}
 
 	// Add a resync period as the PVC's request size can be resized again when we handling
@@ -98,17 +112,55 @@ func NewResizeController(
 		DeleteFunc: ctrl.deletePVC,
 	}, resyncPeriod)
 
+	if handleVolumeInUseError {
+		// list pods so as we can identify PVC that are in-use
+		klog.Infof("Register Pod informer for resizer %s", ctrl.name)
+		podInformer := informerFactory.Core().V1().Pods()
+		ctrl.podLister = podInformer.Lister()
+		ctrl.podListerSynced = podInformer.Informer().HasSynced
+		podInformer.Informer().AddEventHandlerWithResyncPeriod(cache.ResourceEventHandlerFuncs{
+			AddFunc:    ctrl.addPod,
+			DeleteFunc: ctrl.deletePod,
+			UpdateFunc: ctrl.updatePod,
+		}, resyncPeriod)
+	}
+
 	return ctrl
 }
 
 func (ctrl *resizeController) addPVC(obj interface{}) {
-	objKey, err := getPVCKey(obj)
+	objKey, err := getObjectKey(obj)
 	if err != nil {
 		return
 	}
 	ctrl.claimQueue.Add(objKey)
 }
 
+func (ctrl *resizeController) addPod(obj interface{}) {
+	pod := parsePod(obj)
+	if pod == nil {
+		return
+	}
+	ctrl.usedPVCs.addPod(pod)
+}
+
+func (ctrl *resizeController) deletePod(obj interface{}) {
+	pod := parsePod(obj)
+	if pod == nil {
+		return
+	}
+	ctrl.usedPVCs.removePod(pod)
+}
+
+func (ctrl *resizeController) updatePod(oldObj, newObj interface{}) {
+	pod := parsePod(newObj)
+	if pod == nil {
+		return
+	}
+
+	ctrl.usedPVCs.addPod(pod)
+}
+
 func (ctrl *resizeController) updatePVC(oldObj, newObj interface{}) {
 	oldPVC, ok := oldObj.(*v1.PersistentVolumeClaim)
 	if !ok || oldPVC == nil {
@@ -162,14 +214,14 @@ func (ctrl *resizeController) updatePVC(oldObj, newObj interface{}) {
 }
 
 func (ctrl *resizeController) deletePVC(obj interface{}) {
-	objKey, err := getPVCKey(obj)
+	objKey, err := getObjectKey(obj)
 	if err != nil {
 		return
 	}
 	ctrl.claimQueue.Forget(objKey)
 }
 
-func getPVCKey(obj interface{}) (string, error) {
+func getObjectKey(obj interface{}) (string, error) {
 	if unknown, ok := obj.(cache.DeletedFinalStateUnknown); ok && unknown.Obj != nil {
 		obj = unknown.Obj
 	}
@@ -190,9 +242,13 @@ func (ctrl *resizeController) Run(
 	defer klog.Infof("Shutting down external resizer %s", ctrl.name)
 
 	stopCh := ctx.Done()
+	informersSyncd := []cache.InformerSynced{ctrl.pvSynced, ctrl.pvcSynced}
+	if ctrl.handleVolumeInUseError {
+		informersSyncd = append(informersSyncd, ctrl.podListerSynced)
+	}
 
-	if !cache.WaitForCacheSync(stopCh, ctrl.pvSynced, ctrl.pvcSynced) {
-		klog.Errorf("Cannot sync pv/pvc caches")
+	if !cache.WaitForCacheSync(stopCh, informersSyncd...) {
+		klog.Errorf("Cannot sync pod, pv or pvc caches")
 		return
 	}
 
@@ -213,6 +269,7 @@ func (ctrl *resizeController) syncPVCs() {
 
 	if err := ctrl.syncPVC(key.(string)); err != nil {
 		// Put PVC back to the queue so that we can retry later.
+		klog.Errorf("Error syncing PVC: %v", err)
 		ctrl.claimQueue.AddRateLimited(key)
 	} else {
 		ctrl.claimQueue.Forget(key)
@@ -225,18 +282,22 @@ func (ctrl *resizeController) syncPVC(key string) error {
 
 	namespace, name, err := cache.SplitMetaNamespaceKey(key)
 	if err != nil {
-		klog.Errorf("Split meta namespace key of pvc %s failed: %v", key, err)
-		return err
+		return fmt.Errorf("getting namespace and name from key %s failed: %v", key, err)
 	}
 
-	pvc, err := ctrl.pvcLister.PersistentVolumeClaims(namespace).Get(name)
+	pvcObject, exists, err := ctrl.claims.GetByKey(key)
 	if err != nil {
-		if k8serrors.IsNotFound(err) {
-			klog.V(3).Infof("PVC %s/%s is deleted, no need to process it", namespace, name)
-			return nil
-		}
-		klog.Errorf("Get PVC %s/%s failed: %v", namespace, name, err)
-		return err
+		return fmt.Errorf("getting PVC %s/%s failed: %v", namespace, name, err)
+	}
+
+	if !exists {
+		klog.V(3).Infof("PVC %s/%s is deleted or does not exist", namespace, name)
+		return nil
+	}
+
+	pvc, ok := pvcObject.(*v1.PersistentVolumeClaim)
+	if !ok {
+		return fmt.Errorf("expected PVC got: %v", pvcObject)
 	}
 
 	if !ctrl.pvcNeedResize(pvc) {
@@ -244,14 +305,19 @@ func (ctrl *resizeController) syncPVC(key string) error {
 		return nil
 	}
 
-	pv, err := ctrl.pvLister.Get(pvc.Spec.VolumeName)
+	volumeObj, exists, err := ctrl.volumes.GetByKey(pvc.Spec.VolumeName)
 	if err != nil {
-		if k8serrors.IsNotFound(err) {
-			klog.V(3).Infof("PV %s is deleted, no need to process it", pvc.Spec.VolumeName)
-			return nil
-		}
-		klog.Errorf("Get PV %q of pvc %q failed: %v", pvc.Spec.VolumeName, util.PVCKey(pvc), err)
-		return err
+		return fmt.Errorf("Get PV %q of pvc %q failed: %v", pvc.Spec.VolumeName, util.PVCKey(pvc), err)
+	}
+
+	if !exists {
+		klog.Warningf("PV %q bound to PVC %s not found", pvc.Spec.VolumeName, util.PVCKey(pvc))
+		return nil
+	}
+
+	pv, ok := volumeObj.(*v1.PersistentVolume)
+	if !ok {
+		return fmt.Errorf("expected volume but got %+v", volumeObj)
 	}
 
 	if !ctrl.pvNeedResize(pvc, pv) {
@@ -316,12 +382,20 @@ func (ctrl *resizeController) pvNeedResize(pvc *v1.PersistentVolumeClaim, pv *v1
 // 3. Mark pvc as resizing finished(no error, no need to resize fs), need resizing fs or resize failed.
 func (ctrl *resizeController) resizePVC(pvc *v1.PersistentVolumeClaim, pv *v1.PersistentVolume) error {
 	if updatedPVC, err := ctrl.markPVCResizeInProgress(pvc); err != nil {
-		klog.Errorf("Mark pvc %q as resizing failed: %v", util.PVCKey(pvc), err)
-		return err
+		return fmt.Errorf("marking pvc %q as resizing failed: %v", util.PVCKey(pvc), err)
 	} else if updatedPVC != nil {
 		pvc = updatedPVC
 	}
 
+	// if pvc previously failed to expand because it can't be expanded when in-use
+	// we must not try expansion here
+	if ctrl.usedPVCs.hasInUseErrors(pvc) && ctrl.usedPVCs.checkForUse(pvc) {
+		// Record an event to indicate that resizer is not expanding the pvc
+		msg := fmt.Sprintf("Unable to expand %s because CSI driver %s only supports offline expansion and volume is currently in-use", util.PVCKey(pvc), ctrl.resizer.Name())
+		ctrl.eventRecorder.Event(pvc, v1.EventTypeWarning, util.VolumeResizeFailed, msg)
+		return fmt.Errorf(msg)
+	}
+
 	// Record an event to indicate that external resizer is resizing this volume.
 	ctrl.eventRecorder.Event(pvc, v1.EventTypeNormal, util.VolumeResizing,
 		fmt.Sprintf("External resizer is resizing volume %s", pv.Name))
@@ -344,7 +418,6 @@ func (ctrl *resizeController) resizePVC(pvc *v1.PersistentVolumeClaim, pv *v1.Pe
 		// Record an event to indicate that resize operation is failed.
 		ctrl.eventRecorder.Eventf(pvc, v1.EventTypeWarning, util.VolumeResizeFailed, err.Error())
 	}
-
 	return err
 }
 
@@ -352,18 +425,29 @@ func (ctrl *resizeController) resizePVC(pvc *v1.PersistentVolumeClaim, pv *v1.Pe
 func (ctrl *resizeController) resizeVolume(
 	pvc *v1.PersistentVolumeClaim,
 	pv *v1.PersistentVolume) (resource.Quantity, bool, error) {
+
+	// before trying expansion we will remove the PVC from map
+	// that tracks PVCs which can't be expanded when in-use. If
+	// pvc indeed can not be expanded when in-use then it will be added
+	// back when expansion fails with in-use error.
+	ctrl.usedPVCs.removePVCWithInUseError(pvc)
+
 	requestSize := pvc.Spec.Resources.Requests[v1.ResourceStorage]
 
 	newSize, fsResizeRequired, err := ctrl.resizer.Resize(pv, requestSize)
 
 	if err != nil {
-		klog.Errorf("Resize volume %q by resizer %q failed: %v", pv.Name, ctrl.name, err)
-		return newSize, fsResizeRequired, fmt.Errorf("resize volume %s failed: %v", pv.Name, err)
+		// if this error was a in-use error then it must be tracked so as we don't retry without
+		// first verifying if volume is in-use
+		if inUseError(err) {
+			ctrl.usedPVCs.addPVCWithInUseError(pvc)
+		}
+		return newSize, fsResizeRequired, fmt.Errorf("resize volume %q by resizer %q failed: %v", pv.Name, ctrl.name, err)
 	}
 	klog.V(4).Infof("Resize volume succeeded for volume %q, start to update PV's capacity", pv.Name)
 
-	if err := util.UpdatePVCapacity(pv, newSize, ctrl.kubeClient); err != nil {
-		klog.Errorf("Update capacity of PV %q to %s failed: %v", pv.Name, newSize.String(), err)
+	err = ctrl.updatePVCapacity(pv, newSize)
+	if err != nil {
 		return newSize, fsResizeRequired, err
 	}
 	klog.V(4).Infof("Update capacity of PV %q to %s succeeded", pv.Name, newSize.String())
@@ -371,36 +455,6 @@ func (ctrl *resizeController) resizeVolume(
 	return newSize, fsResizeRequired, nil
 }
 
-func (ctrl *resizeController) markPVCResizeInProgress(pvc *v1.PersistentVolumeClaim) (*v1.PersistentVolumeClaim, error) {
-	// Mark PVC as Resize Started
-	progressCondition := v1.PersistentVolumeClaimCondition{
-		Type:               v1.PersistentVolumeClaimResizing,
-		Status:             v1.ConditionTrue,
-		LastTransitionTime: metav1.Now(),
-	}
-	newPVC := pvc.DeepCopy()
-	newPVC.Status.Conditions = util.MergeResizeConditionsOfPVC(newPVC.Status.Conditions,
-		[]v1.PersistentVolumeClaimCondition{progressCondition})
-	return util.PatchPVCStatus(pvc, newPVC, ctrl.kubeClient)
-}
-
-func (ctrl *resizeController) markPVCResizeFinished(
-	pvc *v1.PersistentVolumeClaim,
-	newSize resource.Quantity) error {
-	newPVC := pvc.DeepCopy()
-	newPVC.Status.Capacity[v1.ResourceStorage] = newSize
-	newPVC.Status.Conditions = util.MergeResizeConditionsOfPVC(pvc.Status.Conditions, []v1.PersistentVolumeClaimCondition{})
-	if _, err := util.PatchPVCStatus(pvc, newPVC, ctrl.kubeClient); err != nil {
-		klog.Errorf("Mark PVC %q as resize finished failed: %v", util.PVCKey(pvc), err)
-		return err
-	}
-
-	klog.V(4).Infof("Resize PVC %q finished", util.PVCKey(pvc))
-	ctrl.eventRecorder.Eventf(pvc, v1.EventTypeNormal, util.VolumeResizeSuccess, "Resize volume succeeded")
-
-	return nil
-}
-
 func (ctrl *resizeController) markPVCAsFSResizeRequired(pvc *v1.PersistentVolumeClaim) error {
 	pvcCondition := v1.PersistentVolumeClaimCondition{
 		Type:               v1.PersistentVolumeClaimFileSystemResizePending,
@@ -412,13 +466,132 @@ func (ctrl *resizeController) markPVCAsFSResizeRequired(pvc *v1.PersistentVolume
 	newPVC.Status.Conditions = util.MergeResizeConditionsOfPVC(newPVC.Status.Conditions,
 		[]v1.PersistentVolumeClaimCondition{pvcCondition})
 
-	if _, err := util.PatchPVCStatus(pvc, newPVC, ctrl.kubeClient); err != nil {
-		klog.Errorf("Mark PVC %q as file system resize required failed: %v", util.PVCKey(pvc), err)
-		return err
+	_, err := ctrl.patchClaim(pvc, newPVC)
+
+	if err != nil {
+		return fmt.Errorf("Mark PVC %q as file system resize required failed: %v", util.PVCKey(pvc), err)
 	}
+
 	klog.V(4).Infof("Mark PVC %q as file system resize required", util.PVCKey(pvc))
 	ctrl.eventRecorder.Eventf(pvc, v1.EventTypeNormal,
 		util.FileSystemResizeRequired, "Require file system resize of volume on node")
 
 	return nil
 }
+
+func (ctrl *resizeController) markPVCResizeInProgress(pvc *v1.PersistentVolumeClaim) (*v1.PersistentVolumeClaim, error) {
+	// Mark PVC as Resize Started
+	progressCondition := v1.PersistentVolumeClaimCondition{
+		Type:               v1.PersistentVolumeClaimResizing,
+		Status:             v1.ConditionTrue,
+		LastTransitionTime: metav1.Now(),
+	}
+	newPVC := pvc.DeepCopy()
+	newPVC.Status.Conditions = util.MergeResizeConditionsOfPVC(newPVC.Status.Conditions,
+		[]v1.PersistentVolumeClaimCondition{progressCondition})
+
+	updatedPVC, err := ctrl.patchClaim(pvc, newPVC)
+	if err != nil {
+		return nil, err
+	}
+	return updatedPVC, nil
+}
+
+func (ctrl *resizeController) markPVCResizeFinished(
+	pvc *v1.PersistentVolumeClaim,
+	newSize resource.Quantity) error {
+	newPVC := pvc.DeepCopy()
+	newPVC.Status.Capacity[v1.ResourceStorage] = newSize
+	newPVC.Status.Conditions = util.MergeResizeConditionsOfPVC(pvc.Status.Conditions, []v1.PersistentVolumeClaimCondition{})
+
+	_, err := ctrl.patchClaim(pvc, newPVC)
+	if err != nil {
+		return fmt.Errorf("Mark PVC %q as resize finished failed: %v", util.PVCKey(pvc), err)
+	}
+
+	klog.V(4).Infof("Resize PVC %q finished", util.PVCKey(pvc))
+	ctrl.eventRecorder.Eventf(pvc, v1.EventTypeNormal, util.VolumeResizeSuccess, "Resize volume succeeded")
+
+	return nil
+}
+
+func (ctrl *resizeController) patchClaim(oldPVC, newPVC *v1.PersistentVolumeClaim) (*v1.PersistentVolumeClaim, error) {
+	patchBytes, err := util.GetPVCPatchData(oldPVC, newPVC)
+	if err != nil {
+		return nil, fmt.Errorf("can't patch status of PVC %s as generate path data failed: %v", util.PVCKey(oldPVC), err)
+	}
+	updatedClaim, updateErr := ctrl.kubeClient.CoreV1().PersistentVolumeClaims(oldPVC.Namespace).
+		Patch(context.TODO(), oldPVC.Name, types.StrategicMergePatchType, patchBytes, metav1.PatchOptions{}, "status")
+	if updateErr != nil {
+		return nil, fmt.Errorf("can't patch status of  PVC %s with %v", util.PVCKey(oldPVC), updateErr)
+	}
+	err = ctrl.claims.Update(updatedClaim)
+	if err != nil {
+		return nil, fmt.Errorf("error updating PVC %s in local cache: %v", util.PVCKey(newPVC), err)
+	}
+
+	return updatedClaim, nil
+}
+
+func (ctrl *resizeController) updatePVCapacity(pv *v1.PersistentVolume, newCapacity resource.Quantity) error {
+	klog.V(4).Infof("Resize volume succeeded for volume %q, start to update PV's capacity", pv.Name)
+	newPV := pv.DeepCopy()
+	newPV.Spec.Capacity[v1.ResourceStorage] = newCapacity
+
+	_, err := ctrl.patchPersistentVolume(pv, newPV)
+	if err != nil {
+		return fmt.Errorf("updating capacity of PV %q to %s failed: %v", pv.Name, newCapacity.String(), err)
+	}
+	return nil
+}
+
+func (ctrl *resizeController) patchPersistentVolume(oldPV, newPV *v1.PersistentVolume) (*v1.PersistentVolume, error) {
+	patchBytes, err := util.GetPatchData(oldPV, newPV)
+	if err != nil {
+		return nil, fmt.Errorf("can't update capacity of PV %s as generate path data failed: %v", newPV.Name, err)
+	}
+	updatedPV, updateErr := ctrl.kubeClient.CoreV1().PersistentVolumes().Patch(context.TODO(), newPV.Name, types.StrategicMergePatchType, patchBytes, metav1.PatchOptions{})
+	if updateErr != nil {
+		return nil, fmt.Errorf("update capacity of PV %s failed: %v", newPV.Name, updateErr)
+	}
+	err = ctrl.volumes.Update(updatedPV)
+	if err != nil {
+		return nil, fmt.Errorf("error updating PV %s in local cache: %v", newPV.Name, err)
+	}
+	return updatedPV, nil
+}
+
+func parsePod(obj interface{}) *v1.Pod {
+	if obj == nil {
+		return nil
+	}
+	pod, ok := obj.(*v1.Pod)
+	if !ok {
+		staleObj, ok := obj.(cache.DeletedFinalStateUnknown)
+		if !ok {
+			utilruntime.HandleError(fmt.Errorf("couldn't get stale object %#v", obj))
+			return nil
+		}
+		pod, ok = staleObj.Obj.(*v1.Pod)
+		if !ok {
+			utilruntime.HandleError(fmt.Errorf("stale object is not a Pod %#v", obj))
+			return nil
+		}
+	}
+	return pod
+}
+
+func inUseError(err error) bool {
+	st, ok := status.FromError(err)
+	if !ok {
+		// not a grpc error
+		return false
+	}
+	// if this is a failed precondition error then that means driver does not support expansion
+	// of in-use volumes
+	// More info - https://github.com/container-storage-interface/spec/blob/master/spec.md#controllerexpandvolume-errors
+	if st.Code() == codes.FailedPrecondition {
+		return true
+	}
+	return false
+}
