@@ -1,6 +1,11 @@
 package csi
 
-import "context"
+import (
+	"context"
+	"fmt"
+
+	"github.com/container-storage-interface/spec/lib/go/csi"
+)
 
 func NewMockClient(
 	name string,
@@ -22,7 +27,9 @@ type MockClient struct {
 	supportsControllerResize        bool
 	supportsPluginControllerService bool
 	expandCalled                    int
+	expansionFailed                 bool
 	usedSecrets                     map[string]string
+	usedCapability                  *csi.VolumeCapability
 }
 
 func (c *MockClient) GetDriverName(context.Context) (string, error) {
@@ -41,14 +48,24 @@ func (c *MockClient) SupportsNodeResize(context.Context) (bool, error) {
 	return c.supportsNodeResize, nil
 }
 
+func (c *MockClient) SetExpansionFailed() {
+	c.expansionFailed = true
+}
+
 func (c *MockClient) Expand(
 	ctx context.Context,
 	volumeID string,
 	requestBytes int64,
-	secrets map[string]string) (int64, bool, error) {
+	secrets map[string]string,
+	capability *csi.VolumeCapability) (int64, bool, error) {
 	// TODO: Determine whether the operation succeeds or fails by parameters.
+	if c.expansionFailed {
+		c.expandCalled++
+		return requestBytes, c.supportsNodeResize, fmt.Errorf("expansion failed")
+	}
 	c.expandCalled++
 	c.usedSecrets = secrets
+	c.usedCapability = capability
 	return requestBytes, c.supportsNodeResize, nil
 }
 
@@ -56,6 +73,10 @@ func (c *MockClient) GetExpandCount() int {
 	return c.expandCalled
 }
 
+func (c *MockClient) GetCapability() *csi.VolumeCapability {
+	return c.usedCapability
+}
+
 // GetSecrets returns secrets used for volume expansion
 func (c *MockClient) GetSecrets() map[string]string {
 	return c.usedSecrets
